//--------------------------------------------------------------
// File     : stm32_ub_lcd_ili9341_spi.c
// Datum    : 26.04.2014
// Version  : 1.1
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F4
// IDE      : CooCox CoIDE 1.7.4
// GCC      : 4.7 2012q4
// Module   : GPIO, STM32_UB_SPI2
// Funktion : Grafik-LCD Funktionen (Chip=ILI9341) 
//            Der Zugriff erfolgt ueber SPI (SPI-2)
//            Auflsung : 320 x 240 Pixel
//            Farbmode = 16bit (5R6G5B = RGB565) 
//
// Hinweis  : Das Display benutzt die CPU-Pins :
//            
//            PB11 : CS   (ChipSelect)
//            PB12 : D/C  (DATA/CMD)
//            PB13 : SPI_CLK
//            PB14 : SPI_MISO (MISO / SDO) [nicht notwendig]
//            PB15 : SPI_MOSI (MOSI / SDI)
//
//                  Reset-Pin vom Display an Vcc anschliessen !
//                  Backlight ber einen Widerstand an Vcc
//
// CPU-Interface : IM[3:0]='1110' => 4Wire SPI
//
// SPI-Settings :  SPI-Mode = SPI_MODE_0_MSB, FRQ-Max = 6.6MHz
//                 SPI4 [CLK=PB13, MOSI=PB15, MISO=PB14]
//                 Chip-Select an PB11 
//--------------------------------------------------------------


//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "stm32_ub_lcd_ili9341_spi.h"



//--------------------------------------------------------------
// interne Funktionen
//--------------------------------------------------------------
void P_LCD9341_InitIO(void);
void P_LCD9341_CS(BitAction wert);
void P_LCD9341_WRX(BitAction wert);
void P_LCD9341_InitChip(void);
void P_LCD9341_CMD(uint8_t wert);
void P_LCD9341_DATA(uint8_t wert); 
void P_LCD9341_Delay(volatile uint32_t nCount);
int16_t P_LCD_sgn(int16_t x);




//--------------------------------------------------------------
// Init vom LCD-Display
// Return_wert :
//  -> ERROR   , wenn Display nicht gefunden wurde
//  -> SUCCESS , wenn Display OK
//--------------------------------------------------------------
ErrorStatus UB_LCD_Init(void)
{ 
  ErrorStatus ret_wert=ERROR; 

  // init aller IO-Pins
  P_LCD9341_InitIO();
  // init vom SPI-BUS
  UB_SPI2_Init(SPI_MODE_0);
  // LCD-Controller initialisieren
  P_LCD9341_InitChip();

  ret_wert=SUCCESS;

  LCD_DISPLAY_MODE=PORTRAIT;

  return(ret_wert);
}


//--------------------------------------------------------------
// Fllt den Screen mit einer Farbe
//--------------------------------------------------------------
void UB_LCD_FillScreen(uint16_t color)
{
  uint32_t n;
  uint8_t hi,lo;

  hi=color>>8;
  lo=color&0xFF;

  UB_LCD_SetWindow(0,0,LCD_MAXX-1,LCD_MAXY-1);

  P_LCD9341_CMD(LCD_GRAM);

  for(n=0;n<LCD_PIXEL;n++) {
    P_LCD9341_DATA(hi);
    P_LCD9341_DATA(lo);
  }
}


//--------------------------------------------------------------
// Zeichnet ein Pixel mit einer Farbe an x,y Position
//--------------------------------------------------------------
void UB_LCD_DrawPixel(int16_t xpos, int16_t ypos, uint16_t color)
{
  uint8_t hi,lo;

  UB_LCD_SetWindow(xpos,ypos,xpos,ypos);

  hi=color>>8;
  lo=color&0xFF;

  P_LCD9341_CMD(LCD_GRAM);
  P_LCD9341_DATA(hi);
  P_LCD9341_DATA(lo);

}


//--------------------------------------------------------------
// stellt ein Display-Fenster zum zeichnen ein
// von xtart,ystart zu xend,yend
//--------------------------------------------------------------
void UB_LCD_SetWindow(uint16_t xstart, uint16_t ystart, uint16_t xend, uint16_t yend)
{
  uint8_t hi1,lo1;
  uint8_t hi2,lo2;

  hi1=xstart>>8;
  lo1=xstart&0xFF;

  hi2=xend>>8;
  lo2=xend&0xFF;

  P_LCD9341_CMD(LCD_COLUMN_ADDR);
  P_LCD9341_DATA(hi1);
  P_LCD9341_DATA(lo1);
  P_LCD9341_DATA(hi2);
  P_LCD9341_DATA(lo2);

  hi1=ystart>>8;
  lo1=ystart&0xFF;

  hi2=yend>>8;
  lo2=yend&0xFF;

  P_LCD9341_CMD(LCD_PAGE_ADDR);
  P_LCD9341_DATA(hi1);
  P_LCD9341_DATA(lo1);
  P_LCD9341_DATA(hi2);
  P_LCD9341_DATA(lo2);
}


//--------------------------------------------------------------
// Zeichnet eine Linie mit einer Farbe
// von x1,y1 nach x2,y2
// (benutzt wird der Bresenham-Algorithmus)
//--------------------------------------------------------------
void UB_LCD_DrawLine(int16_t x1, int16_t y1, int16_t x2, int16_t y2, uint16_t color)
{
  int16_t x, y, t, dx, dy, incx, incy, pdx, pdy, ddx, ddy, es, el, err;

  // Entfernung in beiden Dimensionen berechnen
  dx = x2 - x1;
  dy = y2 - y1;

  // Vorzeichen des Inkrements bestimmen
  incx = P_LCD_sgn(dx);
  incy = P_LCD_sgn(dy);
  if(dx<0) dx = -dx;
  if(dy<0) dy = -dy;

  // feststellen, welche Entfernung grer ist
  if (dx>dy) {
    // x ist schnelle Richtung
    pdx=incx; pdy=0;    // pd. ist Parallelschritt
    ddx=incx; ddy=incy; // dd. ist Diagonalschritt
    es =dy;   el =dx;   // Fehlerschritte schnell, langsam
  }
  else {
    // y ist schnelle Richtung
    pdx=0;    pdy=incy; // pd. ist Parallelschritt
    ddx=incx; ddy=incy; // dd. ist Diagonalschritt
    es =dx;   el =dy;   // Fehlerschritte schnell, langsam
  }

  // Initialisierungen vor Schleifenbeginn
  x = x1;
  y = y1;
  err = (el>>1);
  UB_LCD_DrawPixel(x,y,color);

  // Pixel berechnen
  for(t=0; t<el; ++t) // t zaehlt die Pixel, el ist auch Anzahl
  {
    // Aktualisierung Fehlerterm
    err -= es;
    if(err<0) {
      // Fehlerterm wieder positiv (>=0) machen
      err += el;
      // Schritt in langsame Richtung, Diagonalschritt
      x += ddx;
      y += ddy;
    }
    else {
      // Schritt in schnelle Richtung, Parallelschritt
      x += pdx;
      y += pdy;
    }
    UB_LCD_DrawPixel(x,y,color);
  }
}


//--------------------------------------------------------------
// Zeichnet einen Kreis mit einer Farbe
// Mittelpunkt = x1,y1
// Radius = radius
// (benutzt wird der Bresenham-Algorithmus)
//--------------------------------------------------------------
void UB_LCD_DrawCircle(int16_t x0, int16_t y0, int16_t radius, uint16_t color)
{
  int16_t f=1-radius, ddF_x=0, ddF_y=-2*radius, x=0, y=radius;

  UB_LCD_DrawPixel(x0, y0 + radius, color);
  UB_LCD_DrawPixel(x0, y0 - radius, color);
  UB_LCD_DrawPixel(x0 + radius, y0, color);
  UB_LCD_DrawPixel(x0 - radius, y0, color);

  while(x < y)
  {
    if(f >= 0)
    {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x + 1;

    UB_LCD_DrawPixel(x0 + x, y0 + y, color);
    UB_LCD_DrawPixel(x0 - x, y0 + y, color);
    UB_LCD_DrawPixel(x0 + x, y0 - y, color);
    UB_LCD_DrawPixel(x0 - x, y0 - y, color);
    UB_LCD_DrawPixel(x0 + y, y0 + x, color);
    UB_LCD_DrawPixel(x0 - y, y0 + x, color);
    UB_LCD_DrawPixel(x0 + y, y0 - x, color);
    UB_LCD_DrawPixel(x0 - y, y0 - x, color);
  }
}


//--------------------------------------------------------------
// Zeichnet ein Ascii-Zeichen eines Fonts an x,y Position
// mit Vorder- und Hintergrundfarbe (Font = max 16 Pixel breite)
// -> Font muss mit &-Operator uebergeben werden
//--------------------------------------------------------------
void UB_Font_DrawChar(uint16_t x, uint16_t y, uint8_t ascii, UB_Font *font, uint16_t vg, uint16_t bg)
{
  uint16_t xn,yn,start_maske,maske;
  const uint16_t *wert;


  ascii -= 32;
  wert=&font->table[ascii * font->height];

  if(LCD_DISPLAY_MODE==PORTRAIT) {
    start_maske=0x80;
    if(font->width>8) start_maske=0x8000;

    for(yn = 0; yn < font->height; yn++) {
      maske=start_maske;
      // Cursor setzen
      for(xn = 0; xn < font->width; xn++) {
        if((wert[yn] & maske) == 0x00) {
          // Pixel in Hintergrundfarbe zeichnen
          UB_LCD_DrawPixel(x+xn,y+yn,bg);
        }
        else {
          // Pixel in Vordergrundfarbe zeichnen
          UB_LCD_DrawPixel(x+xn,y+yn,vg);
        }
        maske=(maske>>1);
      }
    }
  }
}


//--------------------------------------------------------------
// Zeichnet einen String eines Fonts an x,y Position
// mit Vorder- und Hintergrundfarbe (Font = max 16 Pixel breite)
// -> Font muss mit &-Operator uebergeben werden
//--------------------------------------------------------------
void UB_Font_DrawString(uint16_t x, uint16_t y,char *ptr, UB_Font *font, uint16_t vg, uint16_t bg)
{
  uint16_t pos;

  if(LCD_DISPLAY_MODE==PORTRAIT) {
    pos=x;
    while (*ptr != 0) {
      UB_Font_DrawChar(pos,y,*ptr,font,vg,bg);
      pos+=font->width;
      ptr++;
    }
  }
}


//--------------------------------------------------------------
// interne Funktion
// Init aller IO-Pins fuer das Display
//--------------------------------------------------------------
void P_LCD9341_InitIO(void)
{	
  GPIO_InitTypeDef       GPIO_InitStructure;

  // WRX-Pin
  RCC_AHB1PeriphClockCmd(LCD_WRX_GPIO_CLK, ENABLE);
	  
  GPIO_InitStructure.GPIO_Pin = LCD_WRX_PIN;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_Init(LCD_WRX_GPIO_PORT, &GPIO_InitStructure);

  // ChipSelect-Pin
  RCC_AHB1PeriphClockCmd(LCD_CS_GPIO_CLK, ENABLE);
	  
  GPIO_InitStructure.GPIO_Pin = LCD_CS_PIN;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_Init(LCD_CS_GPIO_PORT, &GPIO_InitStructure);	

  P_LCD9341_CS(Bit_SET);
}


//--------------------------------------------------------------
// interne Funktion
// ChipSelect-Pin schalten
// wert [Bit_SET, Bit_RESET]
//--------------------------------------------------------------
void P_LCD9341_CS(BitAction wert)
{
  if (wert == Bit_RESET) {
    LCD_CS_GPIO_PORT->BSRRH = LCD_CS_PIN;
  }
  else {
    LCD_CS_GPIO_PORT->BSRRL = LCD_CS_PIN;
  }
}


//--------------------------------------------------------------
// interne Funktion
// WRX-Pin schalten
// wert [Bit_SET, Bit_RESET]
//--------------------------------------------------------------
void P_LCD9341_WRX(BitAction wert)
{
  if (wert == Bit_RESET) {
    LCD_WRX_GPIO_PORT->BSRRH = LCD_WRX_PIN;
  }
  else {
    LCD_WRX_GPIO_PORT->BSRRL = LCD_WRX_PIN;
  }
}


//--------------------------------------------------------------
// interne Funktion
// initialisiert den ILI9341-Controller im Display
//--------------------------------------------------------------
void P_LCD9341_InitChip(void)
{
  P_LCD9341_Delay(LCD_INIT_PAUSE);

  P_LCD9341_CMD(LCD_POWERA);
  P_LCD9341_DATA(0x39);
  P_LCD9341_DATA(0x2C);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x34);
  P_LCD9341_DATA(0x02);

  P_LCD9341_CMD(LCD_POWERB);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0xC1);
  P_LCD9341_DATA(0x30);

  P_LCD9341_CMD(LCD_DTCA);
  P_LCD9341_DATA(0x85);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x78);

  P_LCD9341_CMD(LCD_DTCB);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x00);

  P_LCD9341_CMD(LCD_POWER_SEQ);
  P_LCD9341_DATA(0x64);
  P_LCD9341_DATA(0x03);
  P_LCD9341_DATA(0x12);
  P_LCD9341_DATA(0x81);

  P_LCD9341_CMD(LCD_PRC);
  P_LCD9341_DATA(0x20);

  P_LCD9341_CMD(LCD_POWER1);
  P_LCD9341_DATA(0x23);

  P_LCD9341_CMD(LCD_POWER2);
  P_LCD9341_DATA(0x10);

  P_LCD9341_CMD(LCD_VCOM1);
  P_LCD9341_DATA(0x3E);
  P_LCD9341_DATA(0x28);

  P_LCD9341_CMD(LCD_VCOM2);
  P_LCD9341_DATA(0x86);

  P_LCD9341_CMD(LCD_MAC);
  P_LCD9341_DATA(0x48);

  P_LCD9341_CMD(LCD_PIXEL_FORMAT);
  P_LCD9341_DATA(0x55);

  P_LCD9341_CMD(LCD_FRC);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x18);

  P_LCD9341_CMD(LCD_DFC);
  P_LCD9341_DATA(0x08);
  P_LCD9341_DATA(0x82);
  P_LCD9341_DATA(0x27);

  P_LCD9341_CMD(LCD_3GAMMA_EN);
  P_LCD9341_DATA(0x00);

  P_LCD9341_CMD(LCD_COLUMN_ADDR);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0xEF);
  
  P_LCD9341_CMD(LCD_PAGE_ADDR);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x01);
  P_LCD9341_DATA(0x3F);

  P_LCD9341_CMD(LCD_GAMMA);
  P_LCD9341_DATA(0x01);

  P_LCD9341_CMD(LCD_PGAMMA);
  P_LCD9341_DATA(0x0F);
  P_LCD9341_DATA(0x31);
  P_LCD9341_DATA(0x2B);
  P_LCD9341_DATA(0x0C);
  P_LCD9341_DATA(0x0E);
  P_LCD9341_DATA(0x08);
  P_LCD9341_DATA(0x4E);
  P_LCD9341_DATA(0xF1);
  P_LCD9341_DATA(0x37);
  P_LCD9341_DATA(0x07);
  P_LCD9341_DATA(0x10);
  P_LCD9341_DATA(0x03);
  P_LCD9341_DATA(0x0E);
  P_LCD9341_DATA(0x09);
  P_LCD9341_DATA(0x00);

  P_LCD9341_CMD(LCD_NGAMMA);
  P_LCD9341_DATA(0x00);
  P_LCD9341_DATA(0x0E);
  P_LCD9341_DATA(0x14);
  P_LCD9341_DATA(0x03);
  P_LCD9341_DATA(0x11);
  P_LCD9341_DATA(0x07);
  P_LCD9341_DATA(0x31);
  P_LCD9341_DATA(0xC1);
  P_LCD9341_DATA(0x48);
  P_LCD9341_DATA(0x08);
  P_LCD9341_DATA(0x0F);
  P_LCD9341_DATA(0x0C);
  P_LCD9341_DATA(0x31);
  P_LCD9341_DATA(0x36);
  P_LCD9341_DATA(0x0F);

  P_LCD9341_CMD(LCD_SLEEP_OUT);

  P_LCD9341_Delay(LCD_INIT_PAUSE);

  P_LCD9341_CMD(LCD_DISPLAY_ON);

}


//--------------------------------------------------------------
// interne Funktion
// Kommando per SPI an Display senden
//--------------------------------------------------------------
void P_LCD9341_CMD(uint8_t wert)
{
  // Kommando
  P_LCD9341_WRX(Bit_RESET);

  // ChipSelect auf Lo
  P_LCD9341_CS(Bit_RESET);

  // Wert senden
  UB_SPI2_SendByte(wert);
 
  // ChipSelect auf Hi
  P_LCD9341_CS(Bit_SET);
}


//--------------------------------------------------------------
// interne Funktion
// Daten per SPI an Display senden
//--------------------------------------------------------------
void P_LCD9341_DATA(uint8_t wert)
{
  // Data
  P_LCD9341_WRX(Bit_SET);

  // ChipSelect auf Lo
  P_LCD9341_CS(Bit_RESET);

  // Wert senden
  UB_SPI2_SendByte(wert);

  // ChipSelect auf Hi
  P_LCD9341_CS(Bit_SET);
}


//--------------------------------------------------------------
// interne Funktion
// kleine Pause
//-------------------------------------------------------------- 
void P_LCD9341_Delay(volatile uint32_t nCount)
{
  volatile uint32_t index = 0;

  for(index = nCount; index != 0; index--);
}


//--------------------------------------------------------------
// interne Funktion
// Signum funktion
//  Return_Wert
//    1 ,wenn x > 0
//    0 ,wenn x = 0
//   -1 ,wenn x < 0
//--------------------------------------------------------------
int16_t P_LCD_sgn(int16_t x)
{
  return (x > 0) ? 1 : (x < 0) ? -1 : 0;
}

